import Koa from 'koa';
import KoaRouter from 'koa-router';
import koaStatic from 'koa-static';
import compress from 'koa-compress';
import koaBody from 'koa-better-body';
import webpack from 'webpack';
import pug from 'pug';
import fs from 'fs';
import path from 'path';
import CopyWebpackPlugin from 'copy-webpack-plugin';
import {
  koaWebpackDevMiddleware,
  koaWebpackHotMiddleware,
  EntryAssetsWebpackPlugin,
} from 'webpack5-config-react';
import {
  appConfig,
  webpackConfig,
  __DEV__,
  middlewarePath,
  templateDirPath,
} from './config';
import { setRouter } from './utils';

let templateSuffix = 'pug';
let templatePath = appConfig.template;
const entryAssetsName = 'entryAssets.json';
const entryAssetsPath = `${webpackConfig.output?.path}/${entryAssetsName}`;

if (!templatePath) {
  const defaultTemplate = path.resolve(__dirname, './template/index.pug');
  if (fs.existsSync(templateDirPath)) {
    try {
      fs.readdirSync(templateDirPath).forEach(item => {
        if (/\.pug$/.test(item)) {
          templatePath = `${templateDirPath}/${item}`;
          templateSuffix = 'pug';
        } else if (/\.html$/.test(item)) {
          templatePath = `${templateDirPath}/${item}`;
          templateSuffix = 'html';
        }
      });
    } catch (error) {}
  }

  if (!templatePath) {
    templatePath = defaultTemplate;
    templateSuffix = 'pug';
  }
}

webpackConfig.plugins?.push(
  new EntryAssetsWebpackPlugin({
    filename: entryAssetsName,
  })
);
webpackConfig.plugins?.push(
  new CopyWebpackPlugin({
    patterns: [{ from: templatePath, to: webpackConfig.output?.path }],
  })
);

const app = new Koa();
const router = new KoaRouter();
const compiler = webpack(webpackConfig);
const port = appConfig.port || 3000;

__DEV__ &&
  app
    .use(
      koaWebpackDevMiddleware(compiler, {
        publicPath: webpackConfig.output?.publicPath,
      })
    )
    .use(
      koaWebpackHotMiddleware(compiler, {
        log: false,
        path: `/__webpack_hmr`,
        heartbeat: 10 * 1000,
      })
    );

app.use(async (ctx, next) => {
  ctx['appConfig'] = appConfig;
  await next();
});

app.use(
  compress({
    threshold: 2048,
    gzip: {
      flush: require('zlib').Z_SYNC_FLUSH,
    },
  })
);
app.use(koaStatic(webpackConfig.output!.path!));
app.use(koaBody());
/** @description 读取自定义中间件 */
if (fs.existsSync(middlewarePath)) {
  fs.readdirSync(middlewarePath).forEach(item => {
    if (/\.js$/g.test(item)) {
      try {
        const middlewareExport = require(`${middlewarePath}/${item}`);
        const middleware =
          middlewareExport && middlewareExport.default
            ? middlewareExport.default
            : middlewareExport;
        middleware && app.use(middleware);
      } catch (error) {
        // 读取不到则默认跳过
        console.log(error);
      }
    }
  });
}

setRouter(router);

app.use(router.allowedMethods());
app.use(router.routes());

app.use(async (ctx, next) => {
  await next();
  const { status, body } = ctx;
  if (status === 404 && !body) {
    try {
      const assets = fs.existsSync(entryAssetsPath) && require(entryAssetsPath);
      const inject = assets ?? { js: [], css: [] };
      Object.assign(inject, {
        injectData: [],
        ...appConfig,
      });

      ctx.body =
        templateSuffix === 'pug'
          ? pug.renderFile(templatePath, inject)
          : fs.readFileSync(templatePath, { encoding: 'utf8' });
    } catch (error) {
      ctx.status = 500;
      ctx.body = '😭  服务器出错了 😭';
    }
  }
});

app.listen(port, () =>
  console.log(`server running in localhost:${port}`)
).timeout = 60 * 60 * 1000;
